import unittest, sys, os, io, copy
import numpy as np
import cctk

class TestOneIndexedArray(unittest.TestCase):
    def test_indexing(self):
        array = [1,2,3,4,5]
        new_a = cctk.OneIndexedArray(array)

        self.assertEqual(len(new_a), 5)
        self.assertEqual(str(new_a), "[1 2 3 4 5]")
        self.assertEqual(str(new_a[1:3]), "[1 2]")
        self.assertEqual(new_a[5], 5)
        self.assertEqual(new_a[1], 1)
        new_a[1] = 8
        self.assertEqual(str(new_a), "[8 2 3 4 5]")
        self.assertEqual(new_a[1], 8)

        a_2d = [[1,1,1],[2,2,2],[3,3,3]]
        new_a_2d = cctk.OneIndexedArray(a_2d)
        self.assertEqual(new_a_2d[1,1], 1)
        self.assertEqual(new_a_2d[2,0], 2)

        new_a_2d[3,0] = 7
        self.assertEqual(new_a_2d[3,0], 7)

        new_a_2d[3] = np.array([-1, -2, 0])
        self.assertListEqual(list(new_a_2d[3]), [-1, -2, 0])
        self.assertTrue(isinstance(new_a_2d, cctk.OneIndexedArray))

        v = new_a_2d[1]
        self.assertTrue(isinstance(v, np.ndarray))

    def test_smart_indexing(self):
        a1 = cctk.OneIndexedArray([1,2,3])
        a2 = cctk.OneIndexedArray([4,5,6])

        a3 = np.hstack((a1, a2))
        self.assertListEqual(list(a3), [1, 2, 3, 4, 5, 6])
        self.assertTrue(isinstance(a3, np.ndarray))

        a3 = cctk.OneIndexedArray(a3)
        self.assertListEqual(list(a3), [1, 2, 3, 4, 5, 6])
        self.assertTrue(isinstance(a3, cctk.OneIndexedArray))

        a4 = np.vstack((a3,a3))
        a4 = cctk.OneIndexedArray(a4)
        self.assertTrue(a4.shape[0] == 2)
        self.assertTrue(a4.shape[1] == 6)
        self.assertListEqual(list(a4[1]), [1, 2, 3, 4, 5, 6])

        idx = [True, True, False, True, False, False]
        b1 = a3[idx]
        self.assertListEqual(list(b1), [1, 2, 4])
        self.assertTrue(isinstance(b1, cctk.OneIndexedArray))

        idx = [1, 2, 4]
        b2 = a3[idx]
        self.assertListEqual(list(b1), [1, 2, 4])
        self.assertTrue(isinstance(b1, cctk.OneIndexedArray))

if __name__ == '__main__':
    unittest.main()



import unittest, sys, os, io, copy
import numpy as np
import cctk

class TestXYZ(unittest.TestCase):
    def test_readfile(self):
        path = "static/test_peptide.xyz"
        file = cctk.XYZFile.read_file(path)
        self.assertEqual(file.title, "peptide example")

        mol = file.molecule
        self.assertTrue(isinstance(mol, cctk.Molecule))
        self.assertEqual(mol.num_atoms(), 31)
        self.assertTrue(mol.check_for_conflicts())

    def test_writefile(self):
        path = "static/test_peptide.xyz"
        new_path = "static/test_peptide_copy.xyz"

        file = cctk.XYZFile.read_file(path)
        file.write_file(new_path)

        with open(path) as old:
            with open(new_path) as new:
                self.assertListEqual(
                    list(new),
                    list(old)
                )

        os.remove(new_path)

class TestGaussian(unittest.TestCase):
    def test_read_gjf_file(self):
        path = "static/gaussian_file.gjf"
        file = cctk.GaussianFile.read_file(path)
        self.assertEqual(file.header, "#p opt freq=noraman m062x/6-31g(d) scrf=(smd,solvent=diethylether)")
        self.assertEqual(file.title, "title")
        self.assertEqual(file.footer, None)

        mol = file.get_molecule()
        self.assertTrue(isinstance(mol, cctk.Molecule))
        self.assertEqual(mol.num_atoms(), 31)
        self.assertEqual(mol.charge, 0)
        self.assertEqual(mol.multiplicity, 1)

    def test_read_out_file(self):
        path = "static/gaussian_file.out"
        file = cctk.GaussianFile.read_file(path)
        self.assertEqual(file.header, "#p opt freq=noraman m062x/6-31g(d) scrf=(smd,solvent=diethylether)")
        self.assertEqual(file.title, "title")
        self.assertEqual(file.footer, None)
        self.assertTrue(isinstance(file.molecules, cctk.ConformationalEnsemble))

        mol = file.get_molecule()
        self.assertTrue(isinstance(mol, cctk.Molecule))
        self.assertEqual(mol.num_atoms(), 31)
        self.assertEqual(mol.charge, 0)
        self.assertEqual(mol.multiplicity, 1)

        old_path = "static/gaussian_file.gjf"
        new_path = "static/new_gjf.gjf"

        file.write_file(new_path, molecule=2)

        with open(old_path) as old:
            with open(new_path) as new:
                self.assertListEqual(
                    list(new),
                    list(old)
                )

        os.remove(new_path)

class TestMOL2(unittest.TestCase):
    def test_read(self):
        path = "static/dodecane.mol2"
        file = cctk.MOL2File.read_file(path)
        self.assertTrue(isinstance(file, cctk.MOL2File))

        ensemble = file.molecules
        self.assertTrue(isinstance(ensemble, cctk.ConformationalEnsemble))
        self.assertEqual(len(ensemble.molecules), 1)

        mol = ensemble.molecules[0]
        self.assertTrue(isinstance(mol, cctk.Molecule))
        self.assertEqual(len(mol.atomic_numbers), 38)
        self.assertEqual(len(mol.geometry), 38)
        self.assertEqual(mol.get_bond_order(1,2), 1)

    def test_bulk_read(self):
        path = "static/dodecane-csearch.mol2"
        file = cctk.MOL2File.read_file(path)
        self.assertTrue(isinstance(file, cctk.MOL2File))

        ensemble = file.molecules
        self.assertTrue(isinstance(ensemble, cctk.ConformationalEnsemble))
        self.assertEqual(len(ensemble.molecules), 597)
        for mol in ensemble.molecules:
            self.assertEqual(len(mol.atomic_numbers), 38)
            self.assertEqual(len(mol.geometry), 38)
            self.assertEqual(mol.get_bond_order(1,2), 1)

class TestMAE(unittest.TestCase):
    def test_read(self):
        path = "static/dodecane_csearch-out.mae"
        (file, pnames, pvals) = cctk.MAEFile.read_file(path)
        self.assertTrue(isinstance(file, cctk.MAEFile))
        self.assertEqual(len(pnames), 597)
        self.assertEqual(len(pvals), 597)

        ensemble = file.molecules
        self.assertTrue(isinstance(ensemble, cctk.ConformationalEnsemble))
        self.assertEqual(len(ensemble.molecules), 597)
        for mol in ensemble.molecules:
            self.assertEqual(len(mol.atomic_numbers), 38)
            self.assertEqual(len(mol.geometry), 38)
            self.assertEqual(mol.get_bond_order(1,2), 1)

class TestGroup(unittest.TestCase):
    def test_group_add(self):
        path = "static/acetaldehyde.out"
        old_path = "static/14-butanedione.gjf"
        new_path = "static/new_14-butanedione.gjf"

        file = cctk.GaussianFile.read_file(path)
        group = cctk.Group.new_from_molecule(attach_to=6, molecule=file.get_molecule())
        new_mol = cctk.Group.add_group_to_molecule(file.get_molecule(), group, 5)
        file.write_file("static/new_14-butanedione.gjf", molecule=new_mol)

        with open(old_path) as old:
            with open(new_path) as new:
                self.assertListEqual(
                    list(new),
                    list(old)
                )

        os.remove(new_path)

class TestXYZ(unittest.TestCase):
    def test_writefile(self):
        read_path = "static/test_peptide.xyz"
        path = "static/test_peptide.inp"
        new_path = "static/test_peptide_copy.inp"

        file = cctk.XYZFile.read_file(read_path)
        header = "! aug-cc-pVTZ aug-cc-pVTZ/C DLPNO-CCSD(T) TightSCF TightPNO MiniPrint\n%pal nproc 4 end\n%maxcore 4000\n%mdci\n    density none\nend"
        cctk.OrcaFile.write_molecule_to_file(new_path, file.molecule, header)

        with open(path) as old:
            with open(new_path) as new:
                self.assertListEqual(
                    list(new),
                    list(old)
                )

        os.remove(new_path)

        ensemble = cctk.ConformationalEnsemble()
        ensemble.add_molecule(file.molecule)

        orca_file = cctk.OrcaFile(molecules=ensemble, header=header)
        orca_file.write_file(new_path)

        with open(path) as old:
            with open(new_path) as new:
                self.assertListEqual(
                    list(new),
                    list(old)
                )

        os.remove(new_path)

class TestEnsemble(unittest.TestCase):
    def generate_test_ensemble(self):
        path = "static/test_peptide.xyz"
        file = cctk.XYZFile.read_file(path)
        mol = file.molecule

        e1 = np.array([1, 0, 0])
        e2 = np.array([0, 1, 0])
        e3 = np.array([0, 0, 1])

        ensemble = cctk.ConformationalEnsemble()
        ensemble.add_molecule(mol)
        self.assertEqual(len(ensemble.molecules), 1)

        mol_rot = copy.deepcopy(ensemble.molecules[0]).rotate_molecule(e1, 90)
        ensemble.add_molecule(mol_rot)
        self.assertEqual(len(ensemble.molecules), 2)

        mol_trans = copy.deepcopy(ensemble.molecules[0]).translate_molecule(e2)
        ensemble.add_molecule(mol_trans)
        self.assertEqual(len(ensemble.molecules), 3)

        mol_trans_rot = copy.deepcopy(ensemble.molecules[1].translate_molecule(e2))
        ensemble.add_molecule(mol_trans_rot)
        self.assertEqual(len(ensemble.molecules), 4)

        mol_rot_trans = copy.deepcopy(ensemble.molecules[2].rotate_molecule(e1, 90))
        ensemble.add_molecule(mol_rot_trans)
        self.assertEqual(len(ensemble.molecules), 5)

        ensemble.add_molecule(copy.deepcopy(ensemble.molecules[4].rotate_molecule(e3, 20)))
        ensemble.add_molecule(copy.deepcopy(ensemble.molecules[5].rotate_molecule(e3, 20)))
        ensemble.add_molecule(copy.deepcopy(ensemble.molecules[6].rotate_molecule(e3, 20)))
        ensemble.add_molecule(copy.deepcopy(ensemble.molecules[7].rotate_molecule(e3, 20)))
        ensemble.add_molecule(copy.deepcopy(ensemble.molecules[4].rotate_molecule(e3, -20)))
        ensemble.add_molecule(copy.deepcopy(ensemble.molecules[5].rotate_molecule(e3, -20)))
        ensemble.add_molecule(copy.deepcopy(ensemble.molecules[6].rotate_molecule(e3, -20)))
        ensemble.add_molecule(copy.deepcopy(ensemble.molecules[7].rotate_molecule(e3, -20)))
        return ensemble

    def test_align(self):
        #### since all the molecules are identical, every way we do this should be totally fine
        ensemble = self.generate_test_ensemble()
        ensemble = ensemble.align()
        template = ensemble.molecules[0].geometry
        for molecule in ensemble.molecules:
            for i in range(1,len(template)+1):
                self.assertTrue(cctk.helper_functions.compute_distance_between(molecule.geometry[i],template[i]) < 0.0001)

        ensemble2 = self.generate_test_ensemble()
        ensemble2 = ensemble2.align(atoms="heavy")
        template = ensemble2.molecules[0].geometry
        for molecule in ensemble2.molecules:
            for i in range(1,len(template)+1):
                self.assertTrue(cctk.helper_functions.compute_distance_between(molecule.geometry[i],template[i]) < 0.0001)

        ensemble3 = self.generate_test_ensemble()
        ensemble3 = ensemble3.align(atoms=[13, 4, 27, 6, 9, 14])
        template = ensemble3.molecules[0].geometry
        for molecule in ensemble3.molecules:
            for i in range(1,len(template)+1):
                self.assertTrue(cctk.helper_functions.compute_distance_between(molecule.geometry[i],template[i]) < 0.0001)


if __name__ == '__main__':
    unittest.main()



import re
import numpy as np
import networkx as nx

from abc import abstractmethod

from cctk import File, Ensemble, ConformationalEnsemble
from cctk.helper_functions import get_symbol, get_number


class MOL2File(File):
    """
    Generic class for all ``.mol2`` files.

    Attributes:
        name (str): name of file
        molecules (Ensemble): ``Ensemble`` or ``ConformationalEnsemble`` object
    """

    def __init__(self, name=None):
        if isinstance(name, str):
            self.name = name

    @classmethod
    def read_file(cls, filename, name=None, **kwargs):
        """
        Reads ``.mol2`` file and generates a ``MOL2File`` instance.

        Args:
            filename (str): path to file
            name (str): name of the file

        Returns:
            MOL2File object
        """

        file = MOL2File(name=name)

        (geometries, symbols, atom_types, bonds, conformers) = cls._read_mol2(filename, **kwargs)
        atomic_numbers = np.array([get_number(z) for z in symbols], dtype=np.int8)
        atomic_numbers = np.tile(atomic_numbers, (len(geometries), 1))
        if conformers == True:
            file.molecules = ConformationalEnsemble(geometries=geometries, atomic_numbers=atomic_numbers, bonds=[bonds.edges() for g in geometries])
        else:
            file.molecules = Ensemble(geometries=geometries, atomic_numbers=atomic_numbers, bonds=[b.edges() for b in bonds])

        return file

    @classmethod
    def _read_mol2(
        cls, filename, contains_conformers="check", save_memory_for_conformers=True, print_status_messages=False,
    ):
        """
        Reads .mol2 files into cctk.

        Args:
            filename str): the name of the .mol2 file

            contains_conformers('check' or bool): if set to 'check', multiple geometries
                                                in the same file will be compared to see
                                                if they are conformers.  Alternatively,
                                                force the geometries to be treated as
                                                conformers (True) or not (False).  This
                                                latter option increases performance,
                                                particularly for large files.

            save_memory_for_conformers (bool): if True, the first dimension (geometry number) will be
                                            dropped from the symbols and bonds to prevent
                                            the storage of redundant information.  Thus,
                                            symbols will be a one-dimensional :obj:`np.array` of
                                            :obj:`str` and bonds will be a single :obj:`nx.Graph`.

            print_status_messages (bool): if True, update the progerss of the parsing operation to stdout.

        Returns:
            all_geometries, all_clean_symbols, all_symbols, all_bonds, contains_conformers

            all_geometries: np.array(geometry number, atom number, xyz) -> position (float)
            all_clean_symbols: np.array(geometry number, atom number) -> atom symbol (:obj:`str`)
            all_symbols: np.array(geometry number, atom number) -> atom symbol (:obj:`str`)
            all_bonds: list(geometry_number) -> bond connectivity (:obj:`nx.Graph`)
            contains_conformers: bool (True if the geometries correspond to conformers.)
        """
        # read file
        if print_status_messages:
            print(f"Reading {filename}...", end="", flush=True)
        lines = super().read_file(filename)
        if print_status_messages:
            print(f"read {len(lines)} lines...", end="", flush=True)

        # initialize arrays
        all_geometries = []
        all_symbols = []
        all_clean_symbols = []
        all_bonds = []
        this_geometry = []
        this_symbols = []
        this_clean_symbols = []
        this_bonds = None

        # parse file
        i = 0
        in_geometry_block = False
        in_bond_block = False
        bond_number = 0
        while i < len(lines):
            # read the current line
            line = lines[i]

            # determine if we are in a geometry block
            if line.startswith("@<TRIPOS>ATOM"):
                # step forward to the first geometry line
                in_geometry_block = True
                in_bond_block = False
                i += 1
                line = lines[i]
            elif line.startswith("@<TRIPOS>BOND"):
                # update status
                in_geometry_block = False
                in_bond_block = True
                bond_number = 0

                # get next line
                i += 1
                line = lines[i]

                # initialize connectivity graph
                if len(this_geometry) == 0:
                    raise ValueError("got to bond table without a geometry")
                this_bonds = nx.Graph()
                this_bonds.add_nodes_from(range(1, len(this_geometry) + 1))

            # parse geometry if appropriate
            if in_geometry_block:
                fields = re.split(" +", line.strip())
                if len(fields) < 6:
                    print("Error parsing file:")
                    print("Line = '%s'" % line.strip())
                    print(fields)
                    break
                x, y, z = float(fields[2]), float(fields[3]), float(fields[4])
                this_geometry.append([x, y, z])
                symbol = fields[5]
                clean_symbol = fields[1]
                this_symbols.append(symbol)
                this_clean_symbols.append(clean_symbol)
            elif in_bond_block:
                fields = re.split(" +", line.strip())
                if len(fields) == 4:
                    # parse bonds, checking that the bonds are increasing
                    try:
                        this_bond_number = int(fields[0])
                        atom1 = int(fields[1])
                        atom2 = int(fields[2])
                        n_atoms = len(this_geometry)
                        if not 1 <= atom1 <= n_atoms or not 1 <= atom2 <= n_atoms:
                            raise ValueError(f"atom number out of range: {line}")
                        bond_order = int(fields[3])
                        if bond_order <= 0:
                            raise ValueError(f"zero or negative bond order: {line}")
                        if this_bond_number != bond_number + 1:
                            raise ValueError("non-sequential bond number")
                        bond_number = this_bond_number
                        if this_bonds.has_edge(atom1, atom2):
                            current_bond_order = this_bonds[atom1][atom2]["weight"]
                            if current_bond_order != bond_order:
                                raise ValueError(f"inconsistent bond order definition: {line}")
                        this_bonds.add_edge(atom1, atom2, weight=bond_order)
                        this_bonds.add_edge(atom2, atom1, weight=bond_order)
                    except:
                        # assume we have left the bond block
                        in_geometry_block = False
                        in_bond_block = False
                else:
                    # we have left the bond block
                    in_geometry_block = False
                    in_bond_block = False

            # go to next line
            i += 1

            # store geometry and reinitialize if appropriate
            end_of_file = i == len(lines)
            end_of_blocks = not in_geometry_block and not in_bond_block
            if (end_of_file or end_of_blocks) and len(this_geometry) > 0:
                all_geometries.append(this_geometry)
                all_clean_symbols.append(this_clean_symbols)
                all_symbols.append(this_symbols)
                all_bonds.append(this_bonds)
                this_geometry = []
                this_symbols = []
                this_clean_symbols = []
                this_bonds = None

        # convert to numpy array
        all_geometries = np.array(all_geometries)
        all_symbols = np.array(all_symbols)
        all_clean_symbols = np.array(all_clean_symbols)

        # determine if these are conformers
        if contains_conformers == "check":
            contains_conformers = True
            for symbols, bonds in zip(all_symbols[1:], all_bonds[1:]):
                # must have the same symbols and bonds
                if not (all_symbols[0] == symbols).all() or not nx.is_isomorphic(all_bonds[0], bonds):
                    contains_conformers = False
                    break
        elif isinstance(contains_conformers, bool):
            pass
        else:
            raise ValueError("contains_conformers must be 'check' or boolean")

        # if requested, just store one copy of all_symbols and all_bonds
        if save_memory_for_conformers and contains_conformers:
            all_symbols = all_symbols[0]
            all_clean_symbols = all_clean_symbols[0]
            all_bonds = all_bonds[0]

        # return result
        n_geometries = len(all_geometries)
        if print_status_messages:
            if n_geometries > 1:
                if contains_conformers:
                    n_atoms = len(all_geometries[0])
                    n_bonds = all_bonds.number_of_edges()
                    if print_status_messages:
                        print(f"read {n_geometries} conformers ({n_atoms} atoms and {n_bonds} bonds).")
                else:
                    min_n_atoms = len(all_geometries[0])
                    max_n_atoms = len(all_geometries[0])
                    for geometry in all_geometries[1:]:
                        if len(geometry) > max_n_atoms:
                            max_n_atoms = len(geometry)
                        elif len(geometry) < min_n_atoms:
                            min_n_atoms = len(geometry)
                    min_n_bonds = all_bonds[0].number_of_edges()
                    max_n_bonds = all_bonds[0].number_of_edges()
                    for bonds in all_bonds[1:]:
                        if bonds.number_of_edges() > max_n_bonds:
                            max_n_bonds = bonds.number_of_edges()
                        elif bonds.number_of_edges() < min_n_bonds:
                            min_n_bonds = bonds.number_of_edges
                    if print_status_messages:
                        print(f"read {n_geometries} unrelated geometries ({min_n_atoms}-{max_n_atoms} atoms and {min_n_bonds}-{max_n_bonds}) bonds).")
            else:
                n_atoms = len(all_geometries)
                n_bonds = all_bonds.number_of_edges()
                if print_status_messages:
                    print(f"read one geometry ({n_atoms} atoms and {n_bonds} bonds).")

        #### sometimes these labels switch? so gotta check
        if len(all_clean_symbols[0]) < len(all_symbols[0]):
            return (all_geometries, all_clean_symbols, all_symbols, all_bonds, contains_conformers)
        else:
            return (all_geometries, all_symbols, all_clean_symbols, all_bonds, contains_conformers)



import re
import numpy as np
import networkx as nx

from abc import abstractmethod

from cctk import File, Ensemble, ConformationalEnsemble
from cctk.helper_functions import get_symbol, get_number


class MAEFile(File):
    """
    Generic class for all ``.mae`` files.

    Attributes:
        name (str): name of file
        molecules (Ensemble): ``Ensemble`` or ``ConformationalEnsemble`` object
    """

    def __init__(self, name=None):
        if isinstance(name, str):
            self.name = name

    @classmethod
    def read_file(cls, filename, name=None, **kwargs):
        """
        Reads ``.mae`` file and generates a ``MAEFile`` instance.

        Args:
            filename (str): path to file
            name (str): name of the file

        Returns:
            MAEFile object
            property names (list)
            property_values (list)
        """

        file = MAEFile(name=name)

        (geometries, symbols, bonds, p_names, p_vals, conformers) = cls._read_mae(filename, **kwargs)
        atomic_numbers = np.array([get_number(z) for z in symbols], dtype=np.int8)
        atomic_numbers = np.tile(atomic_numbers, (len(geometries), 1))
        if conformers == True:
            file.molecules = ConformationalEnsemble(geometries=geometries, atomic_numbers=atomic_numbers, bonds=[bonds.edges() for g in geometries])
        else:
            file.molecules = Ensemble(geometries=geometries, atomic_numbers=atomic_numbers, bonds=[b.edges() for b in bonds])

        return file, p_names, p_vals

    @classmethod
    def _read_mae(
        cls, filename, contains_conformers="check", save_memory_for_conformers=True, print_status_messages=False,
    ):
        """
        Reads uncompressed Macromodel files.

        Args:
            filename (str): path to file
            contains_conformers (str): one of ``check``, ``True``, or ``False``
            save_memory_for_conformers (Bool):
            print_status_messages (Bool):

        Returns:
            geometries (np.array): array of 3-tuples of geometries
            symbols (np.array): array of atom symbols (str)
            bonds (nx.Graph): ``NetworkX`` graph of bond information
            property_names:
            property_values:
            contains_conformers (Bool): whether or not the file contains conformers
        """
        # read file
        if print_status_messages:
            print(f"Reading {filename}...", end="", flush=True)
        lines = super().read_file(filename)
        if print_status_messages:
            print(f"read {len(lines)} lines...", end="", flush=True)

        # initialize arrays
        geometries = []
        symbols = []
        bonds = []
        property_names = []
        property_values = []
        this_geometry = None
        this_symbols = None
        this_bonds = None
        this_property_names = None
        this_property_values = None

        # parse file
        i = 0
        current_block_type = None
        while i < len(lines):
            # read the current line
            line = lines[i]
            i += 1

            # determine if we are in a molecule block
            end_of_file = i + 1 == len(lines)
            if current_block_type is None and (line.startswith("f_m_ct") or end_of_file):
                # store the current results if any
                if this_geometry is not None and len(this_geometry) > 0:
                    geometries.append(this_geometry)
                    symbols.append(this_symbols)
                    bonds.append(this_bonds)
                    property_names.append(this_property_names)
                    property_values.append(this_property_values)

                # prepare to read a new molecule
                current_block_type = "property_names"
                this_geometry = []
                this_symbols = []
                this_bonds = None
                this_property_names = []
                this_property_values = []
                continue

            # read property names
            elif current_block_type == "property_names":
                line = line.strip()
                if line.startswith("i_m_ct_format"):
                    next_line = lines[i].strip()
                    if next_line != ":::":
                        raise ValueError(f"expected ':::' here but line {i+1} is:\n{next_line}\n")
                    current_block_type = "property_values"
                    i += 1
                elif line.startswith(":::"):
                    raise ValueError(f"expected to see i_m_ct_format as the last property (line {i+1})")
                else:
                    fields = re.split(" +", line)
                    if len(fields) != 1:
                        raise ValueError(f"unexpected number of fields in property name line: {line}")
                    this_property_names.append(line)

            # read property values
            elif current_block_type == "property_values":
                n_properties = len(this_property_names)
                for j in range(n_properties):
                    this_property_values.append(lines[i + j])
                i += n_properties
                current_block_type = "looking_for_geometry1"

            # look for geometry block
            elif current_block_type == "looking_for_geometry1":
                if line.startswith("  m_atom"):
                    current_block_type = "looking_for_geometry2"
            elif current_block_type == "looking_for_geometry2":
                if line.strip() == ":::":
                    current_block_type = "geometry_block"

            # parse geometry
            elif current_block_type == "geometry_block":
                line = line.strip()
                if line == ":::":
                    current_block_type = "bond_block"

                    # initialize bond connectivity graph
                    this_bonds = nx.Graph()
                    n_atoms = len(this_symbols)
                    this_bonds.add_nodes_from(range(1, n_atoms + 1))
                    i += 7
                else:
                    fields = re.split(" +", line)
                    x, y, z = float(fields[2]), float(fields[3]), float(fields[4])
                    this_geometry.append((x, y, z))
                    symbol = fields[-1]
                    this_symbols.append(symbol)

            # parse bonds
            elif current_block_type == "bond_block":
                line = line.strip()
                if line == ":::":
                    current_block_type = None
                else:
                    fields = re.split(" +", line)
                    bond_number, atom1, atom2, bond_order = (
                        int(fields[0]),
                        int(fields[1]),
                        int(fields[2]),
                        int(fields[3]),
                    )
                    n_atoms = len(this_geometry)
                    if not 1 <= atom1 <= n_atoms or not 1 <= atom2 <= n_atoms:
                        raise ValueError(f"atom number out of range: {line}")
                    bond_order = int(fields[3])
                    if bond_order <= 0:
                        raise ValueError(f"zero or negative bond order: {line}")
                    if this_bonds.number_of_edges() != bond_number - 1:
                        raise ValueError(f"non-sequential bond number (expected {this_bonds.number_of_edges()+1} but got {bond_number})")
                    if this_bonds.has_edge(atom1, atom2):
                        current_bond_order = this_bonds[atom1][atom2]["weight"]
                        if current_bond_order != bond_order:
                            raise ValueError(f"inconsistent bond order definition: {line}")
                    this_bonds.add_edge(atom1, atom2, weight=bond_order)
                    this_bonds.add_edge(atom2, atom1, weight=bond_order)

        # convert to numpy array
        geometries = np.array(geometries)
        symbols = np.array(symbols)
        property_names = np.array(property_names)
        property_values = np.array(property_values)

        # determine if these are conformers
        if contains_conformers == "check":
            contains_conformers = True
            for this_symbols, this_bonds in zip(symbols[1:], bonds[1:]):
                # must have the same symbols and bonds
                if not (symbols[0] == this_symbols).all() or not nx.is_isomorphic(bonds[0], this_bonds):
                    contains_conformers = False
                    break
        elif isinstance(contains_conformers, bool):
            pass
        else:
            raise ValueError("contains_conformers must be 'check' or boolean")

        # if requested, just store one copy of symbols and bonds
        if save_memory_for_conformers and contains_conformers:
            symbols = symbols[0]
            bonds = bonds[0]

        # return result
        n_geometries = len(geometries)
        if print_status_messages:
            if n_geometries > 1:
                if contains_conformers:
                    n_atoms = len(geometries[0])
                    n_bonds = bonds.number_of_edges()
                    if print_status_messages:
                        print(f"read {n_geometries} conformers ({n_atoms} atoms and {n_bonds} bonds).")
                else:
                    min_n_atoms = len(geometries[0])
                    max_n_atoms = len(geometries[0])
                    for geometry in geometries[1:]:
                        if len(geometry) > max_n_atoms:
                            max_n_atoms = len(geometry)
                        elif len(geometry) < min_n_atoms:
                            min_n_atoms = len(geometry)
                    min_n_bonds = bonds[0].number_of_edges()
                    max_n_bonds = bonds[0].number_of_edges()
                    for this_bonds in bonds[1:]:
                        if this_bonds.number_of_edges() > max_n_bonds:
                            max_n_bonds = this_bonds.number_of_edges()
                        elif this_bonds.number_of_edges() < min_n_bonds:
                            min_n_bonds = bonds.number_of_edges
                    if print_status_messages:
                        print(f"read {n_geometries} unrelated geometries ({min_n_atoms}-{max_n_atoms} atoms and {min_n_bonds}-{max_n_bonds}) bonds).")
            else:
                n_atoms = len(geometries)
                n_bonds = bonds.number_of_edges()
                if print_status_messages:
                    print(f"read one geometry ({n_atoms} atoms and {n_bonds} bonds).")

        # return result
        return (
            geometries,
            symbols,
            bonds,
            property_names,
            property_values,
            contains_conformers,
        )



import sys
import os
import re
import numpy as np

from abc import ABC, abstractmethod


class File(ABC):
    """
    Represents a text file for use as input or output to another program like Gaussian.
    This class is abstract.
    """

    @abstractmethod
    def __init__(self):
        pass

    @staticmethod
    def write_file(filename, text, overwrite_existing=True):
        """
        Writes output text to a file.

        Args:
            filename (str): path to file, including name (e.g. ``path/to/input.gjf``)
            text (str): desired contents of file
            overwrite_existing (Bool): whether any existing files should be overwritten or not

        Returns:
            ``True`` if write succeeded, ``False`` otherwise
        """
        if not isinstance(text, str):
            raise TypeError("cannot write non-string to file!")
            return False

        if not overwrite_existing and os.path.exists(filename):
            raise ValueError(f"{filename} already exists but not allowed to overwrite")
        else:
            try:
                with open(filename, "w+") as output_file:
                    output_file.write(text)
                return True
            except OSError as e:
                print(e)
                return False

    @staticmethod
    def append_to_file(filename, text):
        """
        Appends output text to a file.

        Args:
            filename (str): path to file, including name (e.g. ``path/to/input.gjf``)
            text (str): desired contents of file

        Returns:
            ``True`` if write succeeded, ``False`` otherwise
        """
        if not isinstance(text, str):
            raise TypeError("cannot write non-string to file!")
            return False

        if os.path.exists(filename):
            try:
                with open(filename, "a+") as output_file:
                    output_file.write(text)
                return True
            except OSError as e:
                print(e)
                return False
        else:
            raise ValueError(f"{filename} does not exist")

    @staticmethod
    def read_file(filename):
        """
        Reads a file and parses into lines.

        Args:
            filename (str): The path to the file.

        Returns:
            A list containing all the lines in the file.
        """
        with open(filename, "r") as filehandle:
            lines = filehandle.read().splitlines()
            return lines



import sys, re
import numpy as np

from abc import abstractmethod

from cctk import File, Molecule, ConformationalEnsemble
from cctk.helper_functions import get_symbol, compute_distance_between, compute_angle_between, compute_dihedral_between, get_number


class OrcaFile(File):
    """
    Generic class for all Orca `.inp` and `.out` files.

    Attributes:
        title (str): the title from the file
        molecules (ConformationalEnsemble): `ConformationalEnsemble` instance
        header (str): file header
    """

    def __init__(self, molecules, title=None, header=None):
        if molecules and isinstance(molecules, ConformationalEnsemble):
            self.molecules = molecules
        if title and (isinstance(title, str)):
            self.title = title
        if header and (isinstance(header, str)):
            self.header = header

    @classmethod
    def read_file(cls, filename):
        pass

    def write_file(self, filename, molecule=None, header=None):
        """
        Write a ``.inp`` file, using object attributes. If no header is specified, the object's header will be used.

        Args:
            filename (str): path to the new file
            molecule (int): which molecule to use -- passed to ``self.get_molecule()``.
                Default is -1 (e.g. the last molecule), but positive integers will select from self.molecules (1-indexed).
                A ``Molecule`` object can also be passed, in which case that molecule will be written to the file.
            header (str): header for new file
        """
        if not isinstance(molecule, Molecule):
            molecule = self.get_molecule(molecule)

        if header is None:
            header = self.header

        self.write_molecule_to_file(filename, molecule, header)

    @classmethod
    def write_molecule_to_file(cls, filename, molecule, header):
        """
        Write a ``.inp``file using the given molecule.

        Args:
            filename (str): path to the new file
            molecule (Molecule): which molecule to use -- a``Molecule`` object.
            header (str): header for new file
        """
        if not isinstance(molecule, Molecule):
            raise TypeError("need a valid molecule to write a file!")

        if (header is None) or (not isinstance(header, str)):
            raise ValueError("can't write a file without a header")

        text = f"{header.strip()}\n\n"

        text += f"* xyz {int(molecule.charge)} {int(molecule.multiplicity)}\n"
        for index, Z in enumerate(molecule.atomic_numbers, start=1):
            line = molecule.get_vector(index)
            text += f"{Z:2d}     {line[0]:.8f}      {line[1]:.8f}      {line[2]:.8f}\n"

        text += "*\n"
        text += "\n"

        #### write the file
        super().write_file(filename, text)

    def get_molecule(self, num=None):
        """
        Returns the last molecule (from an optimization job or other multi-molecule jobs) or the only molecule (from other jobs).

        If ``num`` is specified, returns that job (1-indexed for positive numbers). So ``job.get_molecule(3)`` will return the 3rd element of ``job.molecules``, not the 4th.
        """
        # some methods pass num=None, which overrides setting the default above
        if num is None:
            num = -1

        if not isinstance(num, int):
            raise TypeError("num must be int")

        #### enforce 1-indexing for positive numbers
        if num > 0:
            num += -1

        return self.molecules[num]



import unittest, sys, os, io, copy, math
import numpy as np
import cctk

class TestMolecule(unittest.TestCase):
    def load_molecule(self, path="static/test_peptide.xyz"):
        return cctk.XYZFile.read_file(path).molecule

    def test_basic(self):
        mol = self.load_molecule()
        self.assertListEqual(mol.atomic_numbers.tolist(), [7, 1, 6, 1, 6, 6, 1, 8, 7, 1, 6, 1, 6, 6, 1, 8, 8, 6, 1, 1, 1, 9, 9, 9, 9, 6, 8, 6, 1, 1, 1])

        mol.assign_connectivity()
        edges = [(1, 2), (1, 3), (1, 26), (3, 4), (3, 5), (3, 6), (5, 7), (5, 24), (5, 25), (6, 8), (6, 9), (9, 10), (9, 11), (11, 12), (11, 13), (11, 14), (13, 15), (13, 22), (13, 23), (14, 16), (14, 17), (17, 18), (18, 19), (18, 20), (18, 21), (26, 27), (26, 28), (28, 29), (28, 30), (28, 31)]
        self.assertListEqual(list(mol.bonds.edges()), edges)

    def test_distance(self):
        mol = self.load_molecule()
        self.assertTrue(isinstance(mol, cctk.Molecule))

        self.assertEqual(int(round(mol.get_distance(1,2)*10)), 10)
        self.assertEqual(int(round(mol.get_distance(1,3)*10)), 14)
        self.assertEqual(int(round(mol.get_distance(1,9)*10)), 38)

        mol.set_distance(1, 2, 2.00)

        self.assertEqual(int(round(mol.get_distance(1,2)*10)), 20)
        self.assertEqual(int(round(mol.get_distance(1,3)*10)), 14)
        self.assertEqual(int(round(mol.get_distance(1,9)*10)), 38)

        self.assertTrue(mol.check_for_conflicts())
        mol.set_distance(1, 2, 0.01)
        self.assertRaises(ValueError, mol.check_for_conflicts)

    def test_angle(self):
        mol = self.load_molecule()
        self.assertTrue(isinstance(mol, cctk.Molecule))

        self.assertEqual(int(round(mol.get_angle(1,3,5))), 111)
        self.assertEqual(int(round(mol.get_angle(3,5,7))), 110)
        self.assertEqual(int(round(mol.get_angle(5,7,9))), 64)

        mol.set_angle(1, 3, 5, 120)

        self.assertEqual(int(round(mol.get_angle(1,3,5))), 120)

    def test_dihedral(self):
        mol = self.load_molecule()
        self.assertTrue(isinstance(mol, cctk.Molecule))

        self.assertEqual(int(round(mol.get_dihedral(1,3,5,7))), 60)
        self.assertEqual(int(round(mol.get_dihedral(16,14,17,18))), 11)
        self.assertEqual(int(round(mol.get_dihedral(31,28,1,2))), 148)

        theta = [1, 20, 89, 66, 180, 215, 333]
        for t in theta:
            mol.set_dihedral(1, 3, 5, 7, t)
            self.assertEqual(int(round(mol.get_dihedral(1,3,5,7))), t)

    def test_fragment(self):
        mol = self.load_molecule()
        mol.assign_connectivity()
        (frag1, frag2) = mol._get_bond_fragments(3, 5)
        self.assertEqual(len(frag1), 27)
        self.assertEqual(len(frag2), 4)

        self.assertEqual(len(mol._get_fragment_containing(5)), 31)
        mol.remove_bond(3,5)
        self.assertEqual(len(mol._get_fragment_containing(5)), 4)
        self.assertFalse(mol.are_connected(3,5))
        mol.add_bond(3,5)
        self.assertEqual(len(mol._get_fragment_containing(5)), 31)
        self.assertTrue(mol.are_connected(3,5))

    def test_add_atoms(self):
        mol = cctk.Molecule(np.array([2], dtype=np.int8), [[0, 0, 0]])
        self.assertEqual(mol.num_atoms(), 1)

        mol.add_atom("He", [1, 0, 0])
        self.assertListEqual(mol.atomic_numbers.tolist(), [2, 2])
        self.assertEqual(mol.num_atoms(), 2)

        mol.add_atom("Ar", [3, 0, 0])
        self.assertEqual(mol.num_atoms(), 3)

        mol.add_atom_at_centroid("He", [2, 3])
        self.assertEqual(mol.num_atoms(), 4)
        self.assertListEqual(list(mol.get_vector(4)), [2, 0, 0])

    def test_mass_spec(self):
        mol = cctk.Molecule(np.array([12], dtype=np.int8), [[0, 0, 0]])
        masses, weights = mol.calculate_mass_spectrum()
        self.assertListEqual(list(masses), [23.])
        self.assertListEqual(list(weights), [1.])

    def test_selection(self):
        mol = self.load_molecule()
        self.assertListEqual(mol.get_heavy_atoms(), [1, 3, 5, 6, 8, 9, 11, 13, 14, 16, 17, 18, 22, 23, 24, 25, 26, 27, 28])
        self.assertListEqual(mol.get_atoms_by_symbol("F"), [22, 23, 24, 25])

    def test_translate(self):
        mol = cctk.Molecule(np.array([12], dtype=np.int8), [[0, 0, 0]])

        v = np.array([1.5234,1.231234,-1.77777])
        mol = mol.translate_molecule(v)

        self.assertListEqual(mol.geometry.tolist()[0], list(v))
        self.assertTrue(isinstance(mol.geometry, cctk.OneIndexedArray))

        mol2 = self.load_molecule()
        v2 = np.zeros(shape=3)

        mol2_shift = mol2.translate_molecule(v2)
        self.assertListEqual(mol2.geometry.tolist()[0], mol2_shift.geometry.tolist()[0])

if __name__ == '__main__':
    unittest.main()



import sys, re
import numpy as np

from enum import Enum

from cctk import File, Ensemble, Molecule, ConformationalEnsemble
from cctk.helper_functions import get_symbol, compute_distance_between, compute_angle_between, compute_dihedral_between, get_number

import cctk.parse_gaussian as parse


class JobType(Enum):
    """
    Class to contain allowed Gaussian job types. Not an exhaustive list, but should be fairly comprehensive.

    The value should be the Gaussian keyword, to permit automatic assignment.
    """

    SP = "sp"
    OPT = "opt"
    FREQ = "freq"
    IRC = "irc"
    NMR = "nmr"
    POP = "pop"


class GaussianFile(File):
    """
    Class for Gaussian files. Composes ``Ensemble``.

    Attributes:
        molecules (Ensemble): ``Ensemble`` or ``ConformationalEnsemble`` instance
        job_types (list): list of `job_type` instances
        header (str): optional, header of .gjf file
        footer (str): optional, footer of .gjf file
        success (int): number of successful terminations (should be 1 for an opt, 2 for opt and then freq, 1 for a single point energy, etc)
        theory (dict): contains information from header
        energies (list): list of energies for each cycle
        scf_iterations (list): number of iterations per cycle
        max_displacements (list): list of max displacement values for each cycle
        rms_displacements (list): list of rms displacement values for each cycle
        max_forces (list): list of max force values for each cycle
        rms_forces (list): list of rms force values for each cycle
        gradients (list): list of gradient values for each cycle
        frequencies (list): list of frequencies
        gibbs_free_energy (float): gibbs free energy, from vibrational correction
        enthalpy (float): enthalpy, from vibrational correction
        title (str): optional, title of .gjf file
        conformational_ensemble (Bool): whether or not molecules is an ``Ensemble`` or ``ConformationalEnsemble`` instance
    """

    def __init__(
        self, atomic_numbers, geometries, bonds=None, charges=None, multiplicities=None, job_types=None, theory=None, header=None, footer=None, title="title", conformational_ensemble=False,
    ):
        """
        Create new GaussianFile object.

		Args:
            atomic_numbers (list): list of atomic numbers
            geometries (list): list of lists of 3-tuples of xyz coordinates
            bonds (nx.Graph): Graph object containing connectivity information (1-indexed)
            charges (int): the charge of the molecules
            multiplicities (int): the spin states of the molecules (1 corresponds to singlet, 2 to doublet, 3 to triplet, etc. -- so a multiplicity of 1 is equivalent to S=0)
            job_types (list): list of ``job_type`` instances
            header (str): optional, header of ``.gjf`` file
            footer (str): optional, footer of ``.gjf`` file
            theory (dict): optional, contains information from header
            title (str): optional, title of ``.gjf`` file
		"""

        if header and not isinstance(header, str):
            raise TypeError("header needs to be a string")

        if footer and not isinstance(footer, str):
            raise TypeError("footer needs to be a string")

        if title and not isinstance(title, str):
            raise TypeError("title needs to be a string")

        if job_types is not None:
            if not all(isinstance(job, JobType) for job in job_types):
                raise TypeError(f"invalid job type {job}")

        if charges is None:
            charges = list(np.zeros(shape=len(geometries)))

        if multiplicities is None:
            multiplicities = list(np.ones(shape=len(geometries)))

        if (len(atomic_numbers) > 0) and (len(geometries) > 0):
            arguments = {"atomic_numbers": atomic_numbers, "geometries": geometries, "bonds": bonds, "charges": charges, "multiplicities": multiplicities}
            if conformational_ensemble:
                self.molecules = ConformationalEnsemble(**arguments)
            else:
                try:
                    self.molecules = ConformationalEnsemble(**arguments)
                    conformational_ensemble = True
                except AssertionError as e:
                    self.molecules = Ensemble(**arguments)

        self.header = header
        self.footer = footer
        self.title = title
        self.job_types = job_types

    @classmethod
    def write_molecule_to_file(cls, filename, molecule, header, footer=None, memory=32, cores=16, chk_path=None, title="title", append=False):
        """
        Write a ``.gjf`` file using the given molecule.

        Args:
            filename (str): path to the new file
            molecule (Molecule): which molecule to use -- a``Molecule`` object.
            memory (int): how many GB of memory to request
            header (str): header for new file
            footer (str): footer for new file
            cores (int): how many CPU cores to request
            chk_path (str): path to checkpoint file, if desired
            title (str): title of the file, defaults to "title"
            append (Bool): whether or not to append to file using Link1 specifications
        """
        if not isinstance(molecule, Molecule):
            raise TypeError("need a valid molecule to write a file!")

        if (header is None) or (not isinstance(header, str)):
            raise ValueError("can't write a file without a header")

        #### generate the text
        text = ""
        if append:
            text += "--Link1--\n"
        text += f"%nprocshared={int(cores)}\n"
        text += f"%mem={int(memory)}GB\n"
        if chk_path:
            text += f"%chk={chk_path}\n"

        text += f"{header.strip()}\n\n{title}\n\n"

        text += f"{int(molecule.charge)} {int(molecule.multiplicity)}\n"
        for index, Z in enumerate(molecule.atomic_numbers, start=1):
            line = molecule.get_vector(index)
            text += f"{Z:2d} {line[0]:.8f} {line[1]:.8f} {line[2]:.8f}\n"

        text += "\n"
        if footer is not None:
            text += f"{footer.strip()}\n\n"

        #### write the file
        if append:
            super().append_to_file(filename, text)
        else:
            super().write_file(filename, text)

    def write_file(self, filename, molecule=None, header=None, footer=None, **kwargs):
        """
        Write a ``.gjf`` file, using object attributes. If no header is specified, the object's header/footer will be used.

        Args:
            filename (str): path to the new file
            memory (int): how many GB of memory to request
            cores (int): how many CPU cores to request
            chk_path (str): path to checkpoint file, if desired
            molecule (int): which molecule to use -- passed to ``self.get_molecule()``.
                Default is -1 (e.g. the last molecule), but positive integers will select from self.molecules (1-indexed).
                A ``Molecule`` object can also be passed, in which case that molecule will be written to the file.
            header (str): header for new file
            footer (str): footer for new file
        """
        if not isinstance(molecule, Molecule):
            molecule = self.get_molecule(molecule)

        if header is None:
            header = self.header

        if footer is None:
            footer = self.footer

        self.write_molecule_to_file(filename, molecule, header, footer, **kwargs)

    def num_imaginaries(self):
        """
        Returns the number of imaginary frequencies.
        """
        return len(self.imaginaries())

    def imaginaries(self):
        """
        Returns the imaginary frequencies, rounded to the nearest integer.
        """
        if JobType.FREQ in self.job_types:
            return list(map(int, np.array(self.frequencies)[np.array(self.frequencies) < 0]))
        else:
            raise TypeError("not a frequency job! can't get # imaginary frequencies!")

    @classmethod
    def read_file(cls, filename, return_lines=False):
        """
        Reads a Gaussian``.out`` or ``.gjf`` file and populates the attributes accordingly.
        Only footers from ``opt=modredundant`` can be read automatically --  ``genecep`` custom basis sets, &c must be specified manually.

        Will throw ``ValueError`` if there have been no successful iterations.

        Args:
            filename (str): path to the out file
            return_lines (Bool): whether the lines of the file should be returned
        Returns:
            GaussianFile object
            (optional) the lines of the file
        """
        if re.search("gjf$", filename):
            return cls._read_gjf_file(filename, return_lines)

        lines = super().read_file(filename)
        header = parse.search_for_block(lines, "#p", "----")

        #### automatically assign job types based on header
        job_types = []
        for name, member in JobType.__members__.items():
            if re.search(f" {member.value}", str(header)):
                job_types.append(member)

        #### extract parameters
        success = 0
        for line in lines:
            if line.strip().startswith("Normal termination"):
                success += 1

        (geometries, atom_lists, energies, scf_iterations,) = parse.read_geometries_and_energies(lines)
        geometries = np.array(geometries)
        atomic_numbers = []

        #### convert to right datatype
        for atom_list in atom_lists:
            try:
                atomic_numbers.append(np.array(atom_list, dtype=np.int8))
            except:
                atomic_numbers.append(np.array(list(map(get_number, atom_list)), dtype=np.int8))

        footer = None
        if re.search("modredundant", str(header)):
            footer = parse.search_for_block(lines, "^ The following ModRedundant input section", "^ $", count=1, join="\n")
            footer = "\n".join(list(footer.split("\n"))[1:])  # get rid of the first line
            footer = "\n".join([" ".join(list(filter(None, line.split(" ")))) for line in footer.split("\n")])

        bonds = parse.read_bonds(lines)
        charge = [int(x) for x in parse.find_parameter(lines, "Multiplicity", expected_length=6, which_field=2)]
        multip = [int(x) for x in parse.find_parameter(lines, "Multiplicity", expected_length=6, which_field=5)]

        #### parameters don't get printed every time for opt jobs
        if JobType.OPT in job_types:
            bonds = np.repeat(np.array([bonds[0]]), len(atomic_numbers)-1, axis=0).tolist() + [bonds[-1]]
            charge = list(np.tile(np.array(charge[0]), len(atomic_numbers)-1)) + [charge[-1]]
            multip = list(np.tile(np.array(multip[0]), len(atomic_numbers)-1)) + [multip[-1]]

        f = GaussianFile(atomic_numbers, geometries, bonds, job_types=job_types, charges=charge, multiplicities=multip)
        f.energies = energies
        f.scf_iterations = scf_iterations
        f.header = header
        f.footer = footer
        f.success = success

        #### now for some job-type specific attributes
        if JobType.OPT in job_types:
            f.rms_forces = parse.find_parameter(lines, "RMS\s+Force", expected_length=5, which_field=2)
            f.rms_displacements = parse.find_parameter(lines, "RMS\s+Displacement", expected_length=5, which_field=2)

        if JobType.FREQ in job_types:
            enthalpies = parse.find_parameter(lines, "thermal Enthalpies", expected_length=7, which_field=6)
            if len(enthalpies) == 1:
                f.enthalpy = enthalpies[0]
            elif len(enthalpies) > 1:
                raise ValueError("too many enthalpies found!")

            gibbs_vals = parse.find_parameter(lines, "thermal Free Energies", expected_length=8, which_field=7)
            if len(gibbs_vals) == 1:
                f.gibbs_free_energy = gibbs_vals[0]
            elif len(gibbs_vals) > 1:
                raise ValueError("too many gibbs free energies found!")

            frequencies = []
            try:
                frequencies += parse.find_parameter(lines, "Frequencies", expected_length=5, which_field=2)
                frequencies += parse.find_parameter(lines, "Frequencies", expected_length=5, which_field=3)
                frequencies += parse.find_parameter(lines, "Frequencies", expected_length=5, which_field=4)
                f.frequencies = sorted(frequencies)
            except:
                raise ValueError("error finding frequencies")

        if return_lines:
            return f, lines
        else:
            return f

    @classmethod
    def _read_gjf_file(cls, filename, return_lines=False):
        """
        Reads a Gaussian ``.gjf`` file and populates the attributes accordingly.

        Args:
            filename (str): path to the out file
            return_lines (Bool): whether the lines of the file should be returned
        Returns:
            GaussianFile object
            (optional) the lines of the file
        """
        lines = super().read_file(filename)
        header = None
        footer = None
        header_done = False
        title = None
        charge = None
        multip = None
        in_geom = False
        atomic_numbers = []
        geometries = []

        for idx, line in enumerate(lines):
            if header is None:
                if re.match("#", line):
                    header = line
                    continue

            if (title is None) and (header is not None):
                if header_done:
                    if len(line.strip()) > 0:
                        title = line
                else:
                    if len(line.strip()) > 0:
                        header = header + line
                    else:
                        header_done = True
                continue

            if (title is not None) and (charge is None):
                if len(line.strip()) > 0:
                    pieces = list(filter(None, line.split(" ")))
                    assert len(pieces) == 2, f"can't parse line {line}"

                    charge = int(pieces[0])
                    multip = int(pieces[1])
                    in_geom = True
                    continue

            if in_geom == True:
                if len(line.strip()) == 0:
                    in_geom = False
                else:
                    pieces = list(filter(None, line.split(" ")))
                    assert len(pieces) == 4, f"can't parse line {line}"

                    atomic_numbers.append(pieces[0])
                    geometries.append([pieces[1], pieces[2], pieces[3]])

            if (in_geom == False) and (len(geometries) > 0):
                if footer:
                    footer = footer + "\n" + line
                else:
                    if len(line.strip()) > 0:
                        footer = line

        try:
            atomic_numbers = np.array(atomic_numbers, dtype=np.int8)
        except:
            atomic_numbers = np.array(list(map(get_number, atomic_numbers)), dtype=np.int8)

        geometries = np.array([geometries])

        f = GaussianFile([atomic_numbers], geometries, charges=[charge], multiplicities=[multip])
        f.header = header
        f.footer = footer
        f.title = title

        if return_lines:
            return f, lines
        else:
            return f

    def get_molecule(self, num=None):
        """
        Returns the last molecule (from an optimization job) or the only molecule (from other jobs).

        If ``num`` is specified, returns that job (1-indexed for positive numbers). So ``job.get_molecule(3)`` will return the 3rd element of ``job.molecules``, not the 4th.
        """
        # some methods pass num=None, which overrides setting the default above
        if num is None:
            num = -1

        if not isinstance(num, int):
            raise TypeError("num must be int")

        #### enforce 1-indexing for positive numbers
        if num > 0:
            num += -1

        return self.molecules.molecules[num]

    @classmethod
    def write_ensemble_to_file(cls, filename, ensemble, headers, kwargs):
        """
        Writes an Ensemble to a file using Link1 specification.

        Args:
            filename (str): where to write the file
            ensemble (Ensemble): ``Ensemble`` object to write
            headers (list): headers for each ``write_molecule_to_file`` call
            kwargs (list of dict): arguments for each ``write_molecule_to_file`` call
        """
        for idx, molecule in enumerate(ensemble.molecules):
            if idx == 0:
                cls.write_molecule_to_file(filename, molecule, headers[idx], append=False, **kwargs[idx])
            else:
                cls.write_molecule_to_file(filename, molecule, headers[idx], append=True, **kwargs[idx])




import os
import sys

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

from cctk import MOL2File, Group

from . import data, groups

filenames = [
    "MeH.mol2",
    "EtH.mol2",
    "iPrH.mol2",
    "tBuH.mol2",
    "OH2.mol2",
    "OMeH.mol2",
    "NHAcH.mol2",
    "NH3.mol2",
    "NMe2H.mol2",
    "CF3H.mol2",
    "HCN.mol2",
    "HNO2.mol2",
    "HCO2Me.mol2",
    "FH.mol2",
    "ClH.mol2",
    "BrH.mol2",
    "IH.mol2",
    "SF5H.mol2",
    "SO3HH.mol2",
    "AcH.mol2",
    "CHOH.mol2",
]

names = [
    ["methyl", "Me", "CH3",],
    ["ethyl", "Et", "C2H5",],
    ["isopropyl", "iPr", "iC3H7",],
    ["tert-butyl", "tBu", "tC4H9",],
    ["hydroxy", "OH",],
    ["methoxy", "MeO", "OMe", "CH3O",],
    ["acetamido", "NHAc",],
    ["amino", "NH2",],
    ["dimethylamino", "Me2N", "NMe2",],
    ["trifluoromethyl", "CF3",],
    ["cyano", "CN",],
    ["nitro", "NO2",],
    ["carboxylmethyl", "MeO2C", "CO2Me",],
    ["fluoro", "F",],
    ["chloro", "Cl",],
    ["bromo", "Br",],
    ["iodo", "I",],
    ["pentafluorosulfanyl", "SF5",],
    ["sulfonyl", "SO3H",],
    ["acetyl", "Ac", "COMe",],
    ["formyl", "CHO",],
]


def load_group(name):
    filename = None

    for row in names:
        if name in row:
            filename = filenames[names.index(row)]
            break

    assert filename is not None, f"can't find name {name}!"

    with pkg_resources.path(groups, filename) as file:
        mol = MOL2File.read_file(file).molecules[0]
        mol.assign_connectivity()

        #### every molecule is set so you need to attach to atom 2
        new_group = Group.new_from_molecule(attach_to=2, molecule=mol)
        return new_group



import numpy as np
import copy

class OneIndexedArray(np.ndarray):
    """
    Wrapper for ``np.array`` that's indexed from one, not zero, to store atomic numbers and geometries.
    This only works on 1D or 2D arrays. Additionally, only the first index of a 2D array will be 1-indexed.

    Note that ``enumerate(one_indexed_array)`` will throw ``IndexError`` -- instead, use ``enumerate(one_indexed_array, start=1)``.
    """

    def __new__(cls, obj, **kwargs):
        new = np.array(obj, **kwargs).view(cls)
        return new

    def __getitem__(self, index):
        index = copy.deepcopy(index)
        if isinstance(index, slice):
            new_index = slice(index.start - 1, index.stop - 1, index.step)
            return super().__getitem__(new_index)
        elif isinstance(index, int):
            if index > 0:
                return super().__getitem__(index-1)
            elif index == 0:
                raise IndexError("this is a 1-indexed array: no element 0!")
            elif index < 0:
                return super().__getitem__(index)
        elif (isinstance(index, tuple)) and (len(index) == 2):
            if index[0] is None:
                return super().__getitem__((index[0], index[1]))
            elif index[0] > 0:
                return super().__getitem__((index[0]-1, index[1]))
            elif index[0] == 0:
                raise IndexError("this is a 1-indexed array: no element 0!")
            elif index[0] < 0:
                return super().__getitem__((index[0], index[1]))
        elif (isinstance(index, tuple)) and (len(index) == 1):
            return self.__getitem__(index[0])
        elif isinstance(index, np.ndarray):
            if index.dtype == bool:
                return super().__getitem__(index)
            elif index.ndim == 1:
                index[index >= 1] += -1
                return super().__getitem__(index)
            else:
                index[0][index >= 1] += -1
                return super().__getitem__(index)
        else:
            return super().__getitem__(index)
#            raise IndexError(f"invalid index {index} for OneIndexedArray")

    def __setitem__(self, index, value):
        index = copy.deepcopy(index)
        if isinstance(index, int):
            if index > 0:
                if self.ndim == 1:
                    super().__setitem__(index-1, value)
                elif self.ndim == 2:
                    super().__setitem__(index, value)
                else:
                    raise TypeError("this datatype is only defined for 1D and 2D ndarrays")
            elif index == 0:
                raise IndexError("this is a 1-indexed array: no element 0!")
            elif index < 0:
                super().__setitem__(index, value)
        elif (isinstance(index, tuple)) and (len(index) == 2):
            if index[0] is None:
                super().__setitem__((index[0], index[1]), value)
            elif index[0] > 0:
                super().__setitem__((index[0]-1, index[1]), value)
            elif index[0] == 0:
                raise IndexError("this is a 1-indexed array: no element 0!")
            elif index[0] < 0:
                super().__setitem__((index[0], index[1]), value)
        elif (isinstance(index, tuple)) and (len(index) == 1):
            return self.__setitem__(index[0], value)
        elif isinstance(index, np.ndarray):
            if index.dtype == bool:
                super().__setitem__(index, value)
            elif index.ndim == 1:
                index[index >= 1] += -1
                super().__setitem__(index, value)
            else:
                index[0][index >= 1] += -1
                super().__setitem__(index, value)
        else:
            super().__setitem__(index, value)
#            raise IndexError(f"invalid index {index} for OneIndexedArray")

    def __iter__(self):
        for idx in range(1,len(self)+1):
            yield self.__getitem__(idx)





import sys, re, math
import numpy as np
import networkx as nx

from functools import lru_cache

from cctk import OneIndexedArray
from cctk.helper_functions import (
    get_symbol,
    get_number,
    compute_rotation_matrix,
    compute_distance_between,
    compute_angle_between,
    compute_dihedral_between,
    compute_unit_vector,
    get_covalent_radius,
    get_isotopic_distribution,
)

class Molecule:
    """
    Class that represents a single molecule, abstractly.

    In contrast to typical Python behavior, ``atomic_numbers`` and ``geometry`` are indexed from one, to simplify interfacing with computational chemistry programs.
    This has been done by defining a custom wrapper for ``numpy.ndarray`` called ``cctk.OneIndexedArray``.

    All other datatypes are indexed from 0.

    Attributes:
        name (str): for identification, optional
        atomic_numbers (cctk.OneIndexedArray, dtype=np.int8): list of atomic numbers
        geometry (cctk.OneIndexedArray): list of 3-tuples of xyz coordinates - same ordering as ``atomic_numbers``
        bonds (nx.Graph): Graph object containing connectivity information (1-indexed)
        charge (int): the charge of the molecule
        multiplicity (int): the spin state of the molecule (1 corresponds to singlet, 2 to doublet, 3 to triplet, etc. -- so a multiplicity of 1 is equivalent to S=0)
    """

    def __init__(self, atomic_numbers, geometry, name=None, bonds=None, charge=0, multiplicity=1):
        """
        Create new Molecule object, and assign connectivity if needed.

        ``bonds`` must be a list of edges (i.e. an n x 2 ``numpy`` array).
        """
        if len(atomic_numbers) != len(geometry):
            raise ValueError("length of geometry and atomic_numbers does not match!")

        if not all(isinstance(z, np.int8) for z in atomic_numbers) or atomic_numbers.size == 0:
            raise ValueError("invalid atom list")

        if len(geometry) == 0:
            raise ValueError("invalid geometry list")

        try:
            geometry = np.array(geometry)
            geometry = geometry.astype(float)
        except:
            raise TypeError("geometry cannot be cast to ndarray of floats!")

        if not all(all(isinstance(y, float) for y in x) for x in geometry):
            raise TypeError("each element of self.geometry must be a 3-tuple")

        if name and not isinstance(name, str):
            raise TypeError("name must be a string!")

        for atom in atomic_numbers:
            try:
                get_symbol(atom)
            except ValueError:
                raise ValueError(f"unknwon atomic number {atom}")

        if not isinstance(charge, int):
            try:
                charge = int(charge)
            except:
                raise TypeError("charge must be integer or castable to integer!")

        if not isinstance(multiplicity, int):
            try:
                multiplicity = int(multiplicity)
            except:
                raise TypeError("multiplicity must be positive integer or castable to positive integer")
        assert multiplicity > 0, "multiplicity must be positive"

        self.atomic_numbers = OneIndexedArray(atomic_numbers, dtype=np.int8)
        self.geometry = OneIndexedArray(geometry)

        if bonds:
            for bond in bonds:
                if len(bond) != 2:
                    raise ValueError("while 3-center bonding is possible, it's a no-go in cctk")
                self._check_atom_number(bond[0])
                self._check_atom_number(bond[1])

        self.name = name
        self.multiplicity = multiplicity
        self.charge = charge

        self.bonds = nx.Graph()
        self.bonds.add_nodes_from(range(1, len(atomic_numbers) + 1))
        if bonds:
            for bond in bonds:
                self.add_bond(bond[0], bond[1])

    def assign_connectivity(self, cutoff=0.5):
        """
        Automatically recalculates bonds based on covalent radii. If two atoms are closer than the sum of their covalent radii + 0.5 Angstroms, then they are considered bonded.

        Args:
            cutoff (float): the threshold (in Angstroms) for how close two covalent radii must be to be considered bonded

        Returns:
            self
        """

        for i in range(1, self.num_atoms() + 1):
            for j in range(i + 1, self.num_atoms() + 1):
                distance = self.get_distance(i, j)
                r_i = get_covalent_radius(self.get_atomic_number(i))
                r_j = get_covalent_radius(self.get_atomic_number(j))

                # 0.5 A distance is used by RasMol and Chime (documentation available online) and works well, empirically
                if distance < (r_i + r_j + cutoff):
                    self.add_bond(i, j)
                elif self.get_bond_order(i, j):
                    self.remove_bond(i, j)

        return self

    def check_for_conflicts(self, min_buffer=-1, group1=None, group2=None):
        """
        Automatically checks for conflicts based on covalent radii. If two atoms are closer than the sum of their covalent radii + buffer, then they are considered clashing.
        If `group1` and `group2` are selected, then conflicts will only be evaluated between these two groups of atoms.

        Args:
            min_buffer (float): the threshold (in Angstroms) for how close two covalent radii must be to be considered clashing. -1.0 A is default, for no particular reason.
            group1 (list): atoms to evaluate against `group2` (if `None`, defaults to all atoms)
            group2 (list): atoms to evaluate against `group1` (if `None`, defaults to all atoms)

        Returns:
            True if there are no conflicts, False (+ error) if there are
        """

        if group1 is None:
            group1 = list(range(1, self.num_atoms() + 1))

        if group2 is None:
            group2 = list(range(1, self.num_atoms() + 1))

        for atom in group1 + group2:
            self._check_atom_number(atom)

        for i in group1:
            for j in group2:
                if i == j:
                    continue
                distance = self.get_distance(i, j, check=False)
                r_i = get_covalent_radius(self.get_atomic_number(i))
                r_j = get_covalent_radius(self.get_atomic_number(j))

                # 0.5 A distance is used by RasMol and Chime (documentation available online) and works well, empirically
                if distance < (r_i + r_j + min_buffer):
                    raise ValueError(f"atoms {i} and {j} are too close - distance {distance} A!")
                    return False

        return True

    def add_bond(self, atom1, atom2, bond_order=1):
        """
        Adds a new bond to the bond graph, or updates the existing bond order. Will not throw an error if the bond already exists.

        Args:
            atom1 (int): the number of the first atom
            atom2 (int): the number of the second atom
            bond_order (int): bond order of bond between atom1 and atom2
        """
        self._check_atom_number(atom1)
        self._check_atom_number(atom2)

        assert isinstance(bond_order, int), f"bond order {bond_order} must be an integer"
        assert bond_order >= 0, f"bond order {bond_order} must be positive"

        if self.bonds.has_edge(atom1, atom2):
            if bond_order == 0:
                self.bonds.remove_edge(atom1, atom2)
            else:
                if self.bonds[atom1][atom2]["weight"] != bond_order:
                    self.bonds[atom1][atom2]["weight"] = bond_order
        elif bond_order > 0:
            self.bonds.add_edge(atom1, atom2, weight=bond_order)

    def remove_bond(self, atom1, atom2):
        """
        Alias for ``self.add_bond(atom1, atom2, bond_order=0)`` -- more intuitive nomenclature.
        """
        self.add_bond(atom1, atom2, bond_order=0)

    def _check_atom_number(self, number):
        """
        Helper method which performs quick checks on the validity of a given atom number.
        """
        if not isinstance(number, int):
            raise TypeError("atom number must be integer")

        if number > self.num_atoms():
            raise ValueError(f"atom number {number} too large!")

        if number <= 0:
            raise ValueError(f"atom number {number} invalid: must be a positive integer!")

    def formula(self, return_dict=False):
        """
        Returns the atomic formula.

        If ``return_dict`` is ``True``, then returns a ``dictionary`` with keys elemental symbols and values the number of occurrences.

        For instance, ``water.formula()`` would return ``{'O': 1, 'H': 2}``.

        If ``return_dict`` is ``False``, then returns a stringified version of the formula according to standard rules.

        For instance, ``water.formula()`` would return ``H2O``.

        Args:
            return_dict (Bool): if the method should return a string or a dictionary

        Returns:
            a dictionary or string representing the molecule's formula
        """

        formula_dict = {}
        for atom in self.atomic_numbers:
            symbol = get_symbol(atom)
            if symbol in formula_dict:
                formula_dict[symbol] += 1
            else:
                formula_dict[symbol] = 1
        if return_dict == True:
            return formula_dict
        else:
            formula = ""
            elements = list(formula_dict.keys())

            #### H and C always come first
            if "H" in elements:
                elements.remove("H")
                formula += f"H{formula_dict['H']}"

            if "C" in elements:
                elements.remove("C")
                formula += f"C{formula_dict['C']}"

            for element in sorted(elements):
                formula += f"{element}{formula_dict[element]}"

            return formula

    #### very fast but causes errors sometimes... so i'm commenting this out until further consultation.
    #    @lru_cache(maxsize=32)
    def _get_bond_fragments(self, atom1, atom2, bond_order=1):
        """
        Returns the pieces of a molecule that one would obtain by breaking the bond between two atoms. Will throw ``ValueError`` if the atoms are in a ring.
        Useful for distance/angle/dihedral scans -- one fragment can be moved and the other held constant.

        Args:
            atom1 (int): the number of the first atom
            atom2 (int): the number of the second atom
            bond_order (int): bond order of bond between atom1 and atom2

        Returns:
            fragment1: the list of atoms in fragment 1 (containing atom1)
            fragment2: the list of atoms in fragment 2 (containing atom2)

        """

        self._check_atom_number(atom1)
        self._check_atom_number(atom2)

        if (not isinstance(bond_order, int)) or (bond_order < 0):
            raise ValueError("invalid bond order!")

        if self.bonds.has_edge(atom1, atom2):
            self.bonds.remove_edge(atom1, atom2)

            fragments = nx.connected_components(self.bonds)
            fragment1 = []
            fragment2 = []

            for fragment in fragments:
                if atom1 in fragment:
                    if atom2 in fragment:
                        raise ValueError(f"Atom {atom1} and atom {atom2} are in a ring or otherwise connected!")
                    else:
                        fragment1 = fragment
                if atom2 in fragment:
                    fragment2 = fragment

            self.bonds.add_edge(atom1, atom2, weight=bond_order)
            return list(fragment1), list(fragment2)
        else:
            raise ValueError(f"No bond between atom {atom1} and atom {atom2}!")

    def _get_fragment_containing(self, atom):
        """
        Get the fragment containing the atom with number ``atom``.

        Args:
            atom (int): the number of the atom

        Returns:
            a list of all the atoms in the fragment
        """

        self._check_atom_number(atom)
        fragments = nx.connected_components(self.bonds)

        for fragment in fragments:
            if atom in fragment:
                return list(fragment)
                break

    def set_distance(self, atom1, atom2, distance, move="group"):
        """
        Adjusts the ``atom1`` -- ``atom2`` bond length to be a fixed distance by moving atom2.

        If ``move`` is set to "group", then all atoms bonded to ``atom2`` will also be moved.

        If ``move`` is set to "atom", then only atom2 will be moved.

        Args:
            atom1 (int): the number of the first atom
            atom2 (int): the number of the second atom
            distance (float): distance in Angstroms of the final bond
            move (str): determines how fragment moving is handled

        Returns:
            the Molecule object
        """

        self._check_atom_number(atom1)
        self._check_atom_number(atom2)

        if (not isinstance(distance, float)) or (distance < 0):
            raise ValueError(f"invalid value {distance} for distance!")

        atoms_to_move = []
        if move == "group":
            if self.get_bond_order(atom1, atom2):
                _, atoms_to_move = self._get_bond_fragments(atom1, atom2)
            else:
                atoms_to_move = self._get_fragment_containing(atom2)
        elif move == "atom":
            atoms_to_move = [atom2]
        else:
            raise ValueError(f"Invalid option {move} for parameter 'move'!")

        current_distance = self.get_distance(atom1, atom2)

        v1 = self.get_vector(atom1)
        v2 = self.get_vector(atom2)
        vb = v2 - v1

        if np.linalg.norm(vb) - current_distance > 0.00001:
            raise ValueError(f"Error calculating bond distance!")

        #### move all the atoms
        delta = distance - current_distance
        unitv = compute_unit_vector(vb)
        for atom in atoms_to_move:
            self.geometry[atom] = self.geometry[atom] + (delta * unitv)

        #### check everything worked okay...
        v1f = self.get_vector(atom1)
        v2f = self.get_vector(atom2)
        vbf = v2f - v1f

        if np.linalg.norm(vbf) - distance > 0.001:
            new_dist = np.linalg.norm(vbf)
            raise ValueError(f"Error moving bonds -- new distance is {new_dist:.3f}. Operation failed!")

        return self

    def set_angle(self, atom1, atom2, atom3, angle, move="group"):
        """
        Adjusts the ``atom1`` -- ``atom2`` -- ``atom3`` bond angle to be a fixed value by moving ``atom3``.

        If `move` is set to "group", then all atoms bonded to ``atom3`` will also be moved.

        If `move` is set to "atom", then only ``atom3`` will be moved.

        Args:
            atom1 (int): the number of the first atom
            atom2 (int): the number of the second atom
            atom3 (int): the number of the third atom
            angle (float): final value in degrees of the ``atom1`` -- ``atom2`` -- ``atom3`` angle
            move (str): determines how fragment moving is handled

        Returns:
            the Molecule object
        """

        self._check_atom_number(atom1)
        self._check_atom_number(atom2)
        self._check_atom_number(atom3)

        if self.get_distance(atom1, atom2) < 0.01:
            raise ValueError(f"atom {atom1} and atom {atom2} are too close!")

        if self.get_distance(atom2, atom3) < 0.01:
            raise ValueError(f"atom {atom2} and atom {atom3} are too close!")

        if self.get_distance(atom1, atom3) < 0.01:
            raise ValueError(f"atom {atom1} and atom {atom3} are too close!")

        try:
            angle = float(angle)
        except:
            raise TypeError(f"angle {angle} cannot be converted to float!")

        if (not isinstance(angle, float)) or ((angle < 0) or (angle > 360)):
            raise ValueError(f"invalid value {angle} for angle!")

        atoms_to_move = []
        if move == "group":
            if self.get_bond_order(atom2, atom3):
                _, atoms_to_move = self._get_bond_fragments(atom2, atom3)
            elif self.are_connected(atom2, atom3):
                raise ValueError(
                    f"atom {atom2} and atom {atom3} are connected but not bonded -- cannot adjust angle! try manually removing one or more bonds."
                )
            else:
                atoms_to_move = self._get_fragment_containing(atom3)
        elif move == "atom":
            atoms_to_move = [atom3]
        else:
            raise ValueError(f"Invalid option {move} for parameter 'move'!")

        if atom1 in atoms_to_move:
            raise ValueError(
                f"atom {atom1} and atom {atom3} are connected in multiple ways -- cannot adjust angle! try manually removing one or more bonds."
            )

        current_angle = self.get_angle(atom1, atom2, atom3)
        delta = angle - current_angle

        if np.abs(delta) < 0.001:
            return

        #### now the real work begins...

        #### move everything to place atom2 at the origin
        v2 = self.get_vector(atom2)
        self.translate_molecule(-v2)

        v1 = self.get_vector(atom1)
        v3 = self.get_vector(atom3)

        #### perform the actual rotation
        rot_axis = np.cross(v1, v3)
        rot_matrix = compute_rotation_matrix(rot_axis, delta)
        for atom in atoms_to_move:
            self.geometry[atom] = np.dot(rot_matrix, self.get_vector(atom))

        #### and move it back!
        self.translate_molecule(v2)

        final_angle = self.get_angle(atom1, atom2, atom3)

        #### need to compare cosines to prevent insidious phase difficulties (like 0.00 and 359.99)
        if np.abs(math.cos(math.radians(final_angle)) - math.cos(math.radians(angle))) > 0.001:
            raise ValueError(f"Error rotating atoms -- expected angle {angle}, got {final_angle}  -- operation failed!")

        return self

    def set_dihedral(self, atom1, atom2, atom3, atom4, dihedral, move="group34", check_result=True):
        """
        Adjusts the ``atom1`` -- ``atom2`` -- ``atom3`` -- ``atom4`` dihedral angle to be a fixed value by moving atom 4.

        If ``move`` is set to "atom", then only ``atom4`` will be moved.

        If ``move`` is set to "group4", then all atoms bonded to ``atom4`` will also be moved.

        If ``move`` is set to "group34", then all atoms bonded to ``atom3`` and ``atom4`` will also be moved.

        Args:
            atom1 (int): the number of the first atom
            atom2 (int): the number of the second atom
            atom3 (int): the number of the third atom
            atom4 (int): the number of the fourth atom
            dihedral (float): final value in degrees of the ``atom1`` -- ``atom2`` -- ``atom3`` -- ``atom4`` angle
            move (str): determines how fragment moving is handled
            check_result (Bool): whether the final answer should be checked for correctness

        Returns:
            the Molecule object
        """

        self._check_atom_number(atom1)
        self._check_atom_number(atom2)
        self._check_atom_number(atom3)
        self._check_atom_number(atom4)

        for x in [atom1, atom2, atom3, atom4]:
            for y in [atom1, atom2, atom3, atom4]:
                if x <= y:
                    continue
                else:
                    if self.get_sq_distance(x, y, check=False) < 0.001:
                        raise ValueError(f"atom {x} and atom {y} are too close!")

        try:
            dihedral = float(dihedral)
        except:
            raise TypeError(f"dihedral angle {dihedral} cannot be converted to float!")

        if (not isinstance(dihedral, float)) or ((dihedral < 0) or (dihedral > 360)):
            raise ValueError(f"invalid value {dihedral} for dihedral angle!")

        atoms_to_move = []
        if move == "group34":
            #### add atom3's fragment to atom4
            if self.get_bond_order(atom2, atom3):
                _, atoms_to_move = self._get_bond_fragments(atom2, atom3)
            elif self.are_connected(atom2, atom3):
                raise ValueError(
                    f"atom {atom2} and atom {atom3} are connected but not bonded -- cannot adjust dihedral angle! try manually removing one or more bonds."
                )
            else:
                atoms_to_move = self._get_fragment_containing(atom3)

            #### and make sure atom4 is in there too!
            if atom4 not in atoms_to_move:
                atoms_to_move += self._get_fragment_containing(atom4)
        elif move == "group4":
            if self.get_bond_order(atom3, atom4):
                _, atoms_to_move = self._get_bond_fragments(atom3, atom4)
            elif self.are_connected(atom3, atom4):
                raise ValueError(
                    f"atom {atom3} and atom {atom4} are connected but not bonded -- cannot adjust dihedral angle! try manually removing one or more bonds."
                )
            else:
                atoms_to_move = self._get_fragment_containing(atom4)
        elif move == "atom":
            atoms_to_move = [atom4]
        else:
            raise ValueError(f"Invalid option {move} for parameter 'move'!")

        if atom1 in atoms_to_move:
            raise ValueError(
                f"atom {atom1} and atom {atom4} are connected in multiple ways -- cannot adjust dihedral angle! try manually removing one or more bonds."
            )

        if atom2 in atoms_to_move:
            raise ValueError(
                f"atom {atom2} and atom {atom4} are connected in multiple ways -- cannot adjust dihedral angle! try manually removing one or more bonds."
            )

        if atom4 not in atoms_to_move:
            raise ValueError(f"atom {atom4} is not going to be moved... this operation is doomed to fail!")

        current_dihedral = self.get_dihedral(atom1, atom2, atom3, atom4, check=False)
        delta = (dihedral - current_dihedral) % 360

        if np.abs(delta) < 0.001:
            return

        #### now the real work begins...
        #### move everything to place atom2 at the origin
        v3 = self.get_vector(atom3, check=False)
        self.translate_molecule(-v3)

        #### perform the actual rotation
        rot_matrix = compute_rotation_matrix(-self.get_vector(atom2, check=False), delta)

        for atom in atoms_to_move:
            self.geometry[atom] = np.dot(rot_matrix, self.get_vector(atom, check=False))

        error = self.get_dihedral(atom1, atom2, atom3, atom4, check=False) - dihedral

        #### and move it back!
        self.translate_molecule(v3)

        if check_result:
            final_dihedral = self.get_dihedral(atom1, atom2, atom3, atom4, check=False)

            #### need to compare cosines to prevent insidious phase difficulties (like 0.00 and 359.99)
            #### this will throw ValueError for differences of about 2 degrees
            if np.abs(math.cos(math.radians(final_dihedral)) - math.cos(math.radians(dihedral))) > 0.001:
                raise ValueError(f"Error rotating atoms -- expected dihedral angle {dihedral}, got {final_dihedral}  -- operation failed!")

        return self

    def translate_molecule(self, vector):
        """
        Translates the whole molecule by the given vector.

        Args:
            vector (vector): the vector to translate by

        Returns:
            the Molecule object
        """
        for atom in range(1, self.num_atoms() + 1):
            self.geometry[atom] = self.geometry[atom] + vector

        return self

    def rotate_molecule(self, axis, degrees):
        """
        Rotates the whole molecule around the given axis.

        Args:
            axis (vector): the vector to rotate about
            theta (float): how much to rotate (in degrees)

        Returns:
            the Molecule object
        """
        rot_matrix = compute_rotation_matrix(axis, degrees)

        for atom in range(1, self.num_atoms() + 1):
            self.geometry[atom] = np.dot(rot_matrix, self.geometry[atom])

        return self

    def _recurse_through_formula(self, formula, masses, weights, cutoff=0.0000001, mass_precision=4, weight_precision=8):
        """
        Recurses through a formula and generates m/z isotopic pattern using tail recursion.

        To prevent blowup of memory, fragments with very low abundance are ignored. Masses and weights are also rounded after every step.
        To prevent error accumulation, internal precisions several orders of magnitude lower than the precision of interest should be employed.
        The default values should work nicely for low-res MS applications.

        Args:
            formula (np.array, dtype=np.int8): vector containing atoms left to incorporate
            masses (np.array): list of mass fragments at current iteration
            weights (np.array): relative weights at current iteration
            cutoff (float): cutoff for similarity (masses within ``cutoff`` will be combined)
            mass_precision (int): number of decimal places to store for mass
            weight_precision (int): number of decimal places to store for weight

        Returns:
            masses
            weights
        """
        if np.array_equal(formula, np.zeros(shape=92, dtype=np.int8)):
            return masses[np.argsort(masses)], weights[np.argsort(masses)]

        current_e = np.nonzero(formula)[0][0]
        e_masses, e_weights = get_isotopic_distribution(current_e)

        new_masses = np.zeros(shape=(len(masses)*len(e_masses)))
        new_weights = np.zeros(shape=(len(masses)*len(e_masses)))
        for i in range(len(masses)):
            for j in range(len(e_masses)):
                new_masses[i*len(e_masses)+j] = masses[i] + e_masses[j]
                new_weights[i*len(e_masses)+j] = weights[i] * e_weights[j]

        newer_masses, indices = np.unique(np.round(new_masses, decimals=mass_precision), return_inverse=True)
        newer_weights = np.zeros_like(newer_masses)
        for k in range(len(newer_weights)):
            newer_weights[k] = np.sum(new_weights[np.nonzero(indices == k)])
        newer_weights = np.round(newer_weights, decimals=weight_precision)

        formula[current_e] += -1
        above_cutoff = np.nonzero(newer_weights > cutoff)
        return self._recurse_through_formula(formula, newer_masses[above_cutoff], newer_weights[above_cutoff], cutoff, mass_precision, weight_precision)

    def calculate_mass_spectrum(self, **kwargs):
        """
        Generates list of m/z values based on formula string (e.g. "C10H12")

        Final weights rounded to one decimal point (because of low-res MS).
        """
        form_vec = np.zeros(shape=92, dtype=np.int8)
        for z in self.atomic_numbers:
            form_vec[z - 1] += 1

        masses, weights = self._recurse_through_formula(form_vec, [0], [1], **kwargs)

        new_masses, indices = np.unique(np.round(masses, decimals=1), return_inverse=True)
        new_weights = np.zeros_like(new_masses)
        for k in range(len(new_weights)):
            new_weights[k] = np.sum(weights[np.nonzero(indices == k)])
        new_weights = new_weights / np.max(new_weights)

        return new_masses, new_weights

    def add_atom_at_centroid(self, symbol, atom_numbers, weighted=False):
        """
        Adds atom with symbol ``symbol`` at the centroid of the atoms in ``atom_numbers``.

        If ``weighted`` is ``True``, then the centroid calculation will take into account the atomic numbers of the atoms in question (placing the atom closer to more massive atoms).

        Otherwise, the average is unweighted.

        Args:
            symbol (str): the atomic symbol of the atom to be added
            atom_numbers (list): which atoms to put the new atom between
            weighted (Bool): if the centroid calculation should be weighted (see above)

        Returns:
            the Molecule object
        """

        if (not isinstance(atom_numbers, list)) or (len(atom_numbers) < 2):
            raise TypeError("atom_numbers must be list with at least two elements")

        if not isinstance(symbol, str):
            raise TypeError(f"symbol {symbol} must be a string!")

        coords = [None] * len(atom_numbers)
        weights = [1] * len(atom_numbers)
        for index, atom in enumerate(atom_numbers):
            self._check_atom_number(atom)
            coords[index] = self.get_vector(atom)
            if weighted == True:
                weights[index] = self.atomic_numbers[atom]

        new_coord = list(np.average(coords, weights=weights, axis=0))
        return self.add_atom(coordinates=new_coord, symbol=symbol)

    def add_atom(self, symbol, coordinates):
        """
        Add an atom with symbol ``symbol`` at position ``coordinates``.

        Args:
            symbol (str): symbol of the atom (e.g. "Cl", "Ar", "C")
            coordinates (list): the coordinates to add

        Returns:
            the Molecule object
        """

        if (not isinstance(coordinates, list)) or (len(coordinates) != 3):
            raise TypeError("coordinates must be list with three elements")

        if not isinstance(symbol, str):
            raise TypeError(f"symbol {symbol} must be a string!")

        number = get_number(symbol)
        self.atomic_numbers = np.append(self.atomic_numbers, [number]).view(OneIndexedArray)
        self.geometry = np.append(self.geometry, [coordinates], axis=0).view(OneIndexedArray)
        self.bonds.add_node(self.num_atoms())

        return self

    def remove_atom(self, number):
        """
        Remove the atom with number ``number``.

        Args:
            number (int): number of the atom

        Returns:
            the Molecule object
        """

        self._check_atom_number(number)

        try:
            self.bonds.remove_node(number)
            self.geometry = np.delete(self.geometry, number, axis=0).view(OneIndexedArray)
            self.atomic_numbers = np.delete(self.atomic_numbers, number).view(OneIndexedArray)
            return self
        except:
            raise ValueError("removing atom {number} failed!")

    def get_atomic_number(self, atom):
        """
        Get the atomic number for a given atom.

        Args:
            atom1 (int): number of the first atom

        Returns:
            the atomic number of that atom
        """
        self._check_atom_number(atom)
        return self.atomic_numbers[atom]

    def get_vector(self, atom, atom2=None, check=True):
        """
        Get the geometry vector for a given atom. If two atoms are specified, gives the vector connecting them (from ``atom2`` to ``atom``).
        ``mol.get_vector(atom)`` is thus equivalent to ``mol.get_vector(atom, origin)``.

        Args:
            atom1 (int): number of the first atom
            atom2 (int): number of the second atom (optional)
            check (Bool): whether to validate input data (can be overridden to prevent slow double-checking)

        Returns:
            a Numpy array
        """
        if check:
            self._check_atom_number(atom)

        if atom2:
            if check:
                self._check_atom_number(atom2)
            return (self.geometry[atom] - self.geometry[atom2]).view(np.ndarray)
        else:
            return self.geometry[atom].view(np.ndarray)

    def get_distance(self, atom1, atom2, check=True, _dist=compute_distance_between):
        """
        Wrapper to compute distance between two atoms.

        This function is relatively slow (rate-limiting for certain applications), so performance boosts have been implemented (e.g. preloading ``_dist``).

        Args:
            atom1 (int): number of the first atom
            atom2 (int): number of the second atom
            check (Bool): whether to validate input data (can be overridden to prevent slow double-checking)
            _dist (function): function usd to compute distance

        Returns:
            the distance, in Angstroms
        """
        if check:
            try:
                atom1 = int(atom1)
                atom2 = int(atom2)
            except:
                raise TypeError("atom numbers cannot be cast to int!")

            self._check_atom_number(atom1)
            self._check_atom_number(atom2)

        return _dist(self.get_vector(atom1, check=False), self.get_vector(atom2, check=False))

    def get_sq_distance(self, atom1, atom2, check=True):
        """
        Wrapper to compute squared distance between two atoms -- optimized for speed!

        Args:
            atom1 (int): number of the first atom
            atom2 (int): number of the second atom
            check (Bool): whether to validate input data (can be overridden to prevent slow double-checking)

        Returns:
            the squared distance
        """
        if check:
            try:
                atom1 = int(atom1)
                atom2 = int(atom2)
            except:
                raise TypeError("atom numbers cannot be cast to int!")

            self._check_atom_number(atom1)
            self._check_atom_number(atom2)

        return np.sum(np.square(self.get_vector(atom1, atom2, check=False)))

    def get_angle(self, atom1, atom2, atom3, check=True, _angle=compute_angle_between):
        """
        Wrapper to compute angle between three atoms.

        This function is relatively slow (rate-limiting for certain applications), so performance boosts have been implemented (e.g. preloading ``_angle``).

        Args:
            atom1 (int): number of the first atom
            atom2 (int): number of the second atom
            atom3 (int): number of the third atom
            check (Bool): whether to validate input data (can be overridden to prevent slow double-checking)
            _angle (function): function usd to compute angle

        Returns:
            the angle, in degrees
        """
        if check:
            try:
                atom1 = int(atom1)
                atom2 = int(atom2)
                atom3 = int(atom3)
            except:
                raise TypeError("atom numbers cannot be cast to int!")

            self._check_atom_number(atom1)
            self._check_atom_number(atom2)
            self._check_atom_number(atom3)

        v1 = self.get_vector(atom1, check=False)
        v2 = self.get_vector(atom2, check=False)
        v3 = self.get_vector(atom3, check=False)

        return _angle(v1 - v2, v3 - v2)

    def get_dihedral(self, atom1, atom2, atom3, atom4, check=True, _dihedral=compute_dihedral_between):
        """
        Wrapper to compute dihedral angle between four atoms.

        This function is relatively slow (rate-limiting for certain applications), so performance boosts have been implemented (e.g. preloading ``_dihedral``).

        Args:
            atom1 (int): number of the first atom
            atom2 (int): number of the second atom
            atom3 (int): number of the third atom
            atom4 (int): number of the fourth atom
            check (Bool): whether to validate input data (can be overridden to prevent slow double-checking)
            _dihedral (function): function used to compute dihedral

        Returns:
            the dihedral angle, in degrees
        """
        if check:
            try:
                atom1 = int(atom1)
                atom2 = int(atom2)
                atom3 = int(atom3)
                atom4 = int(atom4)
            except:
                raise TypeError("atom numbers cannot be cast to int!")

            self._check_atom_number(atom1)
            self._check_atom_number(atom2)
            self._check_atom_number(atom3)
            self._check_atom_number(atom4)

        return _dihedral(
            self.get_vector(atom1, check=False),
            self.get_vector(atom2, check=False),
            self.get_vector(atom3, check=False),
            self.get_vector(atom4, check=False),
        )

    def get_bond_order(self, atom1, atom2):
        """
        Wrapper to get bond order between two atoms.

        Args:
            atom1 (int): number of the first atom
            atom2 (int): number of the second atom

        Returns:
            the bond order
        """
        self._check_atom_number(atom1)
        self._check_atom_number(atom2)

        if self.bonds.has_edge(atom1, atom2):
            return self.bonds[atom1][atom2]["weight"]
        else:
            return 0

    def are_connected(self, atom1, atom2):
        """
        Wrapper to tell if two atoms are connected.
        """
        self._check_atom_number(atom1)
        self._check_atom_number(atom2)

        if atom1 in self._get_fragment_containing(atom2):
            return True
        else:
            return False

    def get_atoms_by_symbol(self, symbol):
        """
        Returns all the numbers of atoms of type ``symbol`` in the molecule.
        """
        if not isinstance(symbol, str):
            raise TypeError("symbol {symbol} must be a string")

        number = get_number(symbol)
        atoms = []

        for index, atom in enumerate(self.atomic_numbers, start=1):
            if atom == number:
                atoms.append(index)

        return atoms

    def get_heavy_atoms(self):
        """
        Returns a list of all the heavy atoms in the molecule (i.e., not hydrogen), for array indexing.
        """
        atoms = []

        for index, atom in enumerate(self.atomic_numbers, start=1):
            if atom != 1:
                atoms.append(index)

        return atoms

    def get_adjacent_atoms(self, atom):
        """
        Returns a list of the neighbors of ``atom``. If ``atom`` has no neighbors, an empty list will be returned.
        """
        try:
            atom = int(atom)
        except:
            raise TypeError(f"atom number {atom} cannot be cast to int!")

        self._check_atom_number(atom)

        return list(self.bonds.neighbors(atom))

    def num_atoms(self):
        return len(self.atomic_numbers)

    def rms_distance_between_atoms(self):
        """
        Returns the RMS distance (in Angstroms) between every pair of atoms - a quick, easy-to-calculate proxy for minimizing steric clashes.
        """
        distance = 0
        for i in range(1, self.num_atoms() + 1):
            for j in range(1, self.num_atoms() + 1):
                if i == j:
                    continue
                distance += self.get_distance(i, j) ** 2

        return math.sqrt(distance) / self.num_atoms()

    def optimize_dihedral(self, atom1, atom2, atom3, atom4):
        """
        Minimizes the value of ``self.rms_distance_between_atoms`` for the given dihedral, in one-degree increments.
        A cheap alternative to geometry optimization using *ab initio* methods or density functional theory.

        Args:
            atom1 (int): atom number of the first atom in the dihedral
            atom2 (int): atom number of the second atom in the dihedral
            atom3 (int): atom number of the third atom in the dihedral
            atom4 (int): atom number of the fourth atom in the dihedral

        Returns:
            the final value of the angle
        """
        self._check_atom_number(atom1)
        self._check_atom_number(atom2)
        self._check_atom_number(atom3)
        self._check_atom_number(atom4)

        best_angle = 0
        best_dist = 0

        for angle in range(0, 360):
            self.set_dihedral(atom1, atom2, atom3, atom4, angle)
            if self.rms_distance_between_atoms() > best_dist:
                best_dist = self.rms_distance_between_atoms()
                best_angle = angle

        self.set_dihedral(atom1, atom2, atom3, atom4, best_angle)
        return best_angle

    def atom_string(self, atom):
        """
        Returns the elemental symbol and the atom number for a given atom.

        For example, ``methane.atom_string(1)`` might return "C1".

        Args:
            atom (int): number of the atom

        Returns:
            the aforementioned atom string
        """
        try:
            atom = int(atom)
        except:
            raise ValueError("atom cannot be cast to int")

        self._check_atom_number(atom)

        return f"{get_symbol(self.atomic_numbers[atom])}{atom}"

    def perturb(self, size=0.005):
        """
        This function can be used to generate a slightly different molecule in cases where numerical (or geometric) converge is problematic.

        It adds a random variable (sampled from a normal distribution, centered at 0 with stddev ``size`) to every number in ``self.geometry``.

        Args:
            size (float): stddev of the normal distribution

        Returns:
            the Molecule object
        """
        geometry = self.geometry
        random = np.random.normal(scale=size, size=geometry.shape)

        self.geometry = geometry + random
        return self

    def center(self):
        """
        Moves the centroid to the origin.
        """
        atoms = np.arange(1, self.num_atoms()+1)
        self.translate_molecule(-self.geometry[atoms].mean(axis=0))
        return self

